
/*! yuzu - v2.0.0-rc.16
 * https://github.com/dwightjack/yuzu/packages/yuzu
 * Copyright (c) 2019 - Marco Solazzi;
 * Licensed MIT
 */

import { Events, warn, qs, isElement, bindMethod, nextUid, qsa, evaluate, isPlainObject, noop } from 'yuzu-utils';

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */

function __rest(s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) if (e.indexOf(p[i]) < 0)
            t[p[i]] = s[p[i]];
    return t;
}

const LISTENER_REGEXP = /^([^ ]+)(?: (.+))?$/;
let objDiff = noop;
if (process.env.NODE_ENV !== 'production') {
    objDiff = function objDiff(match, obj, msg) {
        const keys = Object.keys(match);
        const keyStr = keys.length > 0 ? keys.join(', ') : 'no keys';
        Object.keys(obj).forEach((k) => {
            if (keys.indexOf(k) === -1) {
                msg(k, keyStr);
            }
        });
    };
}
/**
 * `Component` is an extensible class constructor which provides the building block of Yuzu component system.
 *
 * It inherits methods from [**Events**](/packages/utils/api/events).
 *
 * ```js
 * const instance = new Component({ ... })
 * ```
 *
 * > **Lifecycle**
 * >
 * > | stage    | hooks     |
 * > |----------| --------- |
 * > | `create` | `created` |
 *
 * @class
 * @param {object} [options={}] Instance options
 * @property {boolean} $active `true` if the instance is mounted and initialized
 * @property {object} options Instance options (see [defaultOptions](#defaultOptions))
 * @property {object} state Instance state (see [setState](#setState))
 * @property {Element} $el The instance root DOM element (see [mount](#mount))
 * @property {Object.<string, Element>} $els  Object mapping references to component's child DOM elements (see `selectors` below)
 * @property {{ string: Component }} $refs Object mapping references to child components (see [setRef](#setRef))
 * @property {object} selectors Object mapping a child element's reference name and a CSS selector or custom function
 * @property {Object.<string, function|string>} listeners Object mapping DOM listeners and handlers (see [setListener](#setListener))
 * @property {Object.<string, function|string>} actions Object mapping state keys and functions to executed on state update
 * @returns {Component}
 */
class Component extends Events {
    // public $$getTree?: IObject;
    /**
     * ```js
     * this.readyState(state, prevState)
     * ```
     *
     * Optional method used to delay the execution of [`ready`](#ready) after the state satisfies to a given condition.
     *
     * Will be executed at every state change until it returns `true`.
     *
     * @name readyState
     * @public
     * @type {function}
     * @param {object} state The current state
     * @param {object} prevState The previous state
     * @returns {boolean}
     */
    /**
     * Component constructor
     */
    constructor(options = {}) {
        super();
        this.$warn = warn(this);
        const defaultOptionsFn = this.constructor
            .defaultOptions;
        let defaultOptions;
        if (typeof defaultOptionsFn === 'function') {
            this.$warn('the static property `defaultOptions` is deprecated. Please move the method to the instance');
            defaultOptions = defaultOptionsFn(this);
        }
        else {
            defaultOptions = this.defaultOptions(this) || {};
        }
        this.state = {};
        if (process.env.NODE_ENV !== 'production') {
            objDiff(defaultOptions, options, (k, keys) => this.$warn(`Option ${k}" has been discarded because it is not defined in component's defaultOptions. Accepted keys are: ${keys}`));
        }
        const entries = Object.entries(defaultOptions);
        this.options = entries.reduce((opts, [key, value]) => {
            let v = options[key] !== undefined ? options[key] : value;
            if (typeof v === 'function' && !Component.isComponent(v)) {
                v = v.bind(this);
            }
            opts[key] = v; // eslint-disable-line no-param-reassign
            return opts;
        }, {});
        this.$active = false;
        // DOM references
        this.$els = {};
        // sub components references
        this.$refs = {};
        this.$refsStore = new Map();
        this.$listeners = new Map();
        this.created();
    }
    /**
     * ```js
     * Component.isComponent(obj)
     * ```
     *
     * Checks whether the passed-in value is a Component constructor.
     *
     * @static
     * @param {*} value
     * @returns {boolean}
     */
    static isComponent(value) {
        return !!(value && value.YUZU_COMPONENT);
    }
    defaultOptions() {
        return {};
    }
    /**
     * ```js
     * mount(el, [state])
     * ```
     *
     * Mounts a component instance on a DOM element and initializes it.
     *
     * ?> To prevent initialization and just mount the component set `state` to `null`.
     *
     * > **Lifecycle**
     * >
     * > | stage    | hooks     |
     * > |----------| --------- |
     * > | `mount` | `beforeMount` <sup>(1)</sup>, `mounted` |
     *
     * 1. Just after attaching the root element (`this.$el`) but before any listener and selector registration.
     *
     * @param {string|Element} el Component's root element
     * @param {object|null} [state={}] Initial state
     * @returns {Component}
     */
    mount(el, state = {}) {
        if (this.$el) {
            throw new Error('Component is already mounted');
        }
        if (this.detached) {
            throw new Error('You cannot mount a detached component. Please use `init` instead');
        }
        const $el = typeof el === 'string' ? qs(el) : el;
        if (!isElement($el)) {
            // fail silently (kinda...);
            this.$warn('Element is not a DOM element', $el);
            return this;
        }
        this.$el = $el; // eslint-disable-line no-multi-assign
        this.beforeMount();
        if (this.selectors) {
            Object.entries(this.selectors).forEach(([key, selector]) => {
                if (typeof selector === 'function') {
                    this.$els[key.replace('[]', '')] = selector(this.$el, this.options);
                    return;
                }
                if (!key.endsWith('[]')) {
                    const el = this.findNode(selector);
                    if (el) {
                        this.$els[key] = el;
                    }
                }
                else {
                    this.$els[key.slice(0, -2)] = this.findNodes(selector);
                }
            });
        }
        if (this.listeners) {
            Object.entries(this.listeners).forEach(([def, handler]) => {
                this.setListener(def, bindMethod(this, handler));
            });
        }
        if (state) {
            this.init(state);
        }
        this.mounted();
        return this;
    }
    /**
     * ```js
     * init([state])
     * ```
     * Initializes the component instance.
     *
     * > **Lifecycle**
     * >
     * > | stage    | hooks     |
     * > |----------| --------- |
     * > | `init` | `initialize`, `ready` |
     *
     * @param {object|null} [state={}] Initial state
     * @returns {Component}
     */
    init(state = {}) {
        if (!this.detached && !isElement(this.$el)) {
            throw new Error('component instance not mounted');
        }
        const { $el } = this;
        // initialization placeholder
        let uid = $el && $el.getAttribute(Component.UID_DATA_ATTR);
        if (uid) {
            this.$warn(`Element ${uid} is already initialized... skipping`, $el);
            this.$uid = uid;
            return this;
        }
        uid = nextUid();
        this.$uid = uid;
        if ($el) {
            $el.setAttribute(Component.UID_DATA_ATTR, uid);
            if (!$el.id) {
                $el.id = `c_${uid}`;
            }
        }
        this.initialize();
        if (this.actions) {
            Object.entries(this.actions).forEach(([key, method]) => {
                this.on(`change:${key}`, bindMethod(this, method));
            });
        }
        const initialState = Object.assign(this.state, state);
        this.replaceState(initialState);
        this.$active = true;
        if (this.readyState) {
            // is it a promise ?
            const watcher = (current, prev) => {
                if (this.readyState(current, prev)) {
                    this.off('change:*', watcher);
                    this.ready();
                }
            };
            this.on('change:*', watcher);
            return this;
        }
        this.ready();
        return this;
    }
    /**
     * Lifecycle hook called on instance creation.
     *
     * At this stage just the instance options (`this.options`) are initialized.
     *
     *  Overwrite this method with custom logic in your components.
     *
     * ?> Use this hook to tap as early as possible into the component's properties. For example to set a dynamic initial state.
     */
    created() { }
    /**
     * Lifecycle hook called just before mounting the instance onto the root element.
     *
     * At this stage the component is already attached to its root DOM element.
     *
     * Overwrite this method with custom logic in your components.
     */
    beforeMount() { }
    /**
     * Lifecycle hook called when the instance gets mounted onto a DOM element.
     *
     * At this stage both children elements (`this.$els.*`) and event listeners configured in `this.listeners` have been setup.
     *
     * Overwrite this method with custom logic in your components.
     *
     * ?> Use this method when you need to work with the DOM or manage any side-effect that requires the component to be into the DOM.
     */
    mounted() { }
    /**
     * Lifecycle hook called before instance initialization.
     *
     * At this stage the state and state listeners are not yet been initialized.
     *
     * Overwrite this method with custom logic in your components.
     *
     * ?> Use this method to set child components by [setRef](#setRef) and run any preparatory work on the instance.
     */
    initialize() { }
    /**
     * Lifecycle hook called after instance initialization.
     *
     * At this stage State and event binding are already in place.
     *
     * Overwrite this method with custom logic in your components.
     *
     * ?> `ready` lifecycle can be delayed (_async ready_) by implementing a [`readyState`](packages/yuzu/#async-ready-state) method.
     */
    ready() { }
    /**
     * Lifecycle hook called just before closing child refs.
     *
     * This hook is called just before destroying the instance. Every property, listener and state feature is still active.
     *
     * Overwrite this method with custom logic in your components.
     *
     * !> This is an async method. Return a promise in order to suspend the destroy process.
     */
    beforeDestroy() { }
    /**
     * Returns an array of elements matching a CSS selector in the context of the component's root element
     *
     * @param selector {string} CSS selector to match
     * @return {Element[]}
     */
    findNodes(selector) {
        return qsa(selector, this.$el);
    }
    /**
     * Returns the first element matching a CSS selector in the context of the component's root element
     *
     * @param selector {string} CSS selector to match
     * @return {Element}
     */
    findNode(selector) {
        return qs(selector, this.$el);
    }
    /**
     * ```js
     * getState(key)
     * ```
     *
     * Returns a property of the state or a default value if the property is not set.
     *
     * ?> In ES6 environments you can use a [destructuring assignment](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Destructuring_assignment) instead: ` const { name = 'John'} = this.state`
     *
     *
     * @param {string} key Property to return
     * @param {*} def Default value
     * @returns {*}
     * @example
     * const instance = new Component().mount('#app', { a: 1 });
     * // instance.getState('a') === 1
     *
     * // instance.getState('b', false) === false
     */
    getState(key, def) {
        return this.state.hasOwnProperty(key) ? this.state[key] : def;
    }
    /* eslint-disable class-methods-use-this */
    /**
     * ```js
     * shouldUpdateState(string, currentValue, newValue)
     * ```
     * Executes a strict inequality comparison (`!==`) on the passed-in values and returns the result.
     *
     * !> This method is executed on every [`setState`](#setstate) call.
     *
     * You can overwrite this method with your own validation logic.
     *
     * @param {string} key State property name
     * @param {*} currentValue Value stored in the current state
     * @param {*} newValue New value
     * @returns {boolean}
     */
    shouldUpdateState(key, currentValue, newValue) {
        return currentValue !== newValue;
    }
    /* eslint-enable class-methods-use-this */
    /**
     * ```js
     * setState(updater, [silent])
     * ```
     *
     * Updates the internal instance state by creating a shallow copy of the current state and updating the passed-in keys.
     *
     * If the computed new state is different from the old one it emits a `change:<property>` event for every changed property
     * as well as a global `change:*` event.
     *
     * If the new value argument is a function, it will be executed with the current state as argument.
     * The returned value will be used to update the state.
     *
     * ?> To prevent an instance from emitting `change:` events set the second argument to `true` (silent update).
     *
     * @param {object|function} updater Defines which part of the state must be updated
     * @param {boolean} [silent=false] Update the state without emitting change events
     * @example
     * instance.on('change:a', (next, prev) => console.log(next, prev));
     * instance.setState({ a: 1 }); //emits 'change:a' -> logs undefined,1
     *
     * instance.setState({ a: 1 }); //nothing happens
     * instance.setState({ a: 2 }, true); //nothing happens, again...
     *
     * // use the current state to calculate its next value
     * instance.setState(({ a }) => ({a + 1}));
     */
    setState(updater, silent = false) {
        const changed = [];
        const { state: prevState } = this;
        const changeSet = evaluate(updater, this.state);
        if (process.env.NODE_ENV !== 'production') {
            objDiff(this.state, changeSet, (k, keys) => `setState: key "${k}" has been discarded because it is not defined in the component's initial state. Accepted keys are: ${keys}`);
        }
        const entries = Object.entries(this.state);
        this.state = entries.reduce((newState, [k, prevValue]) => {
            const value = changeSet[k];
            if (value === undefined ||
                this.shouldUpdateState(k, prevValue, value) === false) {
                newState[k] = prevValue; // eslint-disable-line no-param-reassign
            }
            else {
                changed.push(k);
                newState[k] = value; // eslint-disable-line no-param-reassign
            }
            return newState;
        }, {});
        if (!silent && changed.length > 0) {
            while (changed.length > 0) {
                const k = changed.pop();
                this.emit(`change:${k}`, this.state[k], prevState[k]);
            }
            this.emit('change:*', this.state, prevState);
        }
    }
    /**
     * ```js
     * replaceState(newState, [silent])
     * ```
     *
     * Replaces the current state of the instance with a completely new state.
     *
     * !> Note that this methods is un-affected by [`shouldUpdateState`](#shouldupdatestate).
     *
     * @param {object} newState The new state object
     * @param {boolean} [silent=false] Replace the state without emitting change events
     * @example
     * instance.replaceState({ a: 1 });
     * // instance.state.a === 1
     * instance.replaceState({ b: 2 });
     * // instance.state.b === 2
     * // instance.state.a === undefined
     */
    replaceState(newState, silent = false) {
        const { state: prevState } = this;
        this.state = Object.assign({}, newState);
        const entries = Object.entries(this.state);
        entries.forEach(([key, value]) => {
            if (!silent) {
                this.emit(`change:${key}`, value, prevState[key]);
            }
        });
        if (!silent) {
            this.emit('change:*', this.state, prevState);
        }
    }
    /**
     * ```js
     * broadcast(event, [...params])
     * ```
     *
     * Emits a `broadcast:<eventname>` event on every child component listed in `$refs`.
     *
     * @param {string} event Event name
     * @param {*[]} [params] Additional arguments to pass to the handler
     * @example
     * const child = new Component('#child');
     * child.on('broadcast:log', (str) => console.log(str));
     *
     * instance.setRef({ id: 'child', component: child });
     * instance.broadcast('log', 'test') // child component logs 'test'
     */
    broadcast(event, ...params) {
        this.$refsStore.forEach((instance) => {
            instance.emit(`broadcast:${event}`, ...params);
        });
    }
    /**
     * ```js
     * setListener(string, handler)
     * ```
     *
     * Sets a DOM event listener.
     *
     * The first argument must be a string composed by an event name (ie `click`) and a CSS selector (`.element`)
     * separated by a space.
     *
     * If the CSS selector starts with `@` the listener will be attached to the
     * corresponding reference child element (`this.$els.<element>`), if any.
     *
     * @param {string} def Event and target element definition. Format `eventName [target]`
     * @param {function} handler Event handler
     * @example
     *
     * // attach a click handler to a child element
     * instance.setListener('click .button', () => ...)
     *
     * // attach a click handler to this.$els.btn
     * instance.setListener('click @btn', () => ...)
     *
     * // attach a click handler to this.$el
     * instance.setListener('click', () => ...)
     */
    setListener(def, handler) {
        let event;
        let selector;
        const match = def && def.match(LISTENER_REGEXP);
        if (match) {
            const $el = this.$el;
            [, event, selector = $el] = match;
            let element;
            if (typeof selector === 'string') {
                element = selector.startsWith('@')
                    ? this.$els[selector.slice(1)]
                    : qs(selector, $el);
            }
            else {
                element = selector;
            }
            if (element) {
                if (Array.isArray(element)) {
                    element.forEach((el, i) => {
                        const h = (e) => {
                            handler.call(this, e, i);
                        };
                        el.addEventListener(event, h);
                        this.$listeners.set(h, { event, element: el });
                    });
                }
                else {
                    element.addEventListener(event, handler);
                    this.$listeners.set(handler, { event, element });
                }
            }
        }
    }
    /**
     * ```js
     * removeListeners()
     * ```
     *
     * Removes all DOM event listeners attached with `.setListener`.
     *
     */
    removeListeners() {
        this.$listeners.forEach(({ event, element }, handler) => {
            element.removeEventListener(event, handler);
        });
        this.$listeners.clear();
    }
    /**
     * ```js
     * setRef(config, [props])
     * ```
     *
     * Attaches a reference to a child component.
     *
     * If a reference `id` is already attached, the previous one is destroyed and replaced with the new one.
     *
     * ?> This is an async method returning a promise.
     *
     * @param {object} config A child component configuration object
     * @param {string} config.id Reference id. Will be used to set a reference to the child component onto `this.$refs`
     * @param {component} config.component Component constructor or component instance
     * @param {string|HTMLElement} [config.el] Child component root element. This property is ignored if `config.component` is a component instance or a detached component constructor
     * @param {Object} [config.on] Child component event listeners. Format `{ 'eventname': handler }`
     * @param {*} config.* Any other property listed here will be passed to the constructor as option
     * @param {object} [props] Child component initial state
     * @returns {Promise}
     * @example
     * const parent = new Component('#root');
     *
     * class ChildComponent extends Component {}
     *
     * // as constructor
     * parent.setRef({
     *   id: 'child',
     *   component: ChildComponent,
     *   el: '#child',
     *   // other options here...
     * });
     *
     * // as instance
     * parent.setRef({
     *   id: 'child',
     *   component: new ChildComponent({ ... }).mount('#child', null) // <-- prevent component init
     * });
     *
     * // sync parent - child state
     * // updates child `parentCount` state whenever `parent.state.count` changes
     * parent.setRef({
     *   id: 'child',
     *   component: ChildComponent,
     *   el: '#child',
     *   // other options here...
     * }, {
     *   parentCount: (parentState) => parentState.count
     * });
     */
    async setRef(refCfg, props) {
        if (!isPlainObject(refCfg)) {
            throw new TypeError('Invalid reference configuration');
        }
        const { component: ChildComponent, el, id, on } = refCfg, options = __rest(refCfg, ["component", "el", "id", "on"]);
        const { detached } = this;
        let ref;
        // if (el && detached) {
        //   throw new Error(
        //     `setRef "${id}": you cannot define a component with DOM root as child of a detached component.`,
        //   );
        // }
        if (Component.isComponent(ChildComponent)) {
            ref = new ChildComponent(options);
        }
        else if (ChildComponent instanceof Component) {
            ref = ChildComponent;
        }
        else if (typeof ChildComponent === 'function') {
            ref = ChildComponent(this.$el, this.state);
        }
        else {
            throw new TypeError('Invalid reference configuration');
        }
        // $context gets propagated to every child component
        // this way we can use `connect`-ed components as reference
        if (this.$context) {
            Object.defineProperty(ref, '$context', {
                enumerable: false,
                get: () => this.$context,
            });
        }
        Object.defineProperty(ref, '$parent', {
            writable: true,
            enumerable: false,
            value: this,
        });
        if (!id) {
            throw new Error('Invalid reference id string');
        }
        // bind events...
        if (on) {
            Object.entries(on).forEach(([name, handler]) => {
                ref.on(name, handler);
            });
        }
        const { $refs } = this;
        const prevRef = $refs[id];
        const refState = {};
        $refs[id] = ref;
        this.$refsStore.set(id, ref);
        if (!ref.detached && !ref.$el) {
            if (!el) {
                throw new Error(`You need to provide a root element for the child element with id "${id}".`);
            }
            ref.mount(el, null);
        }
        const stateMap = evaluate(props, ref, this);
        if (stateMap) {
            Object.entries(stateMap).forEach(([name, value]) => {
                if (typeof value === 'function') {
                    let key = name;
                    let src = '*';
                    if (name.includes('>')) {
                        [src = '*', key] = name.split('>');
                    }
                    refState[key] = value(src !== '*' ? this.state[src] : this.state, ref);
                    this.on(`change:${src}`, (state) => {
                        ref.setState({ [key]: value(state, ref) });
                    });
                }
                else {
                    refState[name] = value;
                }
            });
        }
        let { $el } = this;
        if (detached && ref.$el) {
            let { $parent } = this;
            while (!$el && $parent) {
                $el = $parent.$el;
                $parent = $parent.$parent;
            }
        }
        if (!$el && ref.$el) {
            throw new Error(`You cannot attach a plain Component to a DetachedComponents tree. (${id})`);
        }
        if (prevRef) {
            await prevRef.destroy();
            if (prevRef.$el &&
                $el.contains(prevRef.$el) &&
                prevRef.$el.parentElement &&
                ref.$el) {
                prevRef.$el.parentElement.replaceChild(ref.$el, prevRef.$el);
            }
            else if (ref.$el && !$el.contains(ref.$el)) {
                $el.appendChild(ref.$el);
            }
            return ref.init(refState);
        }
        if ($el && ref.$el && !$el.contains(ref.$el)) {
            $el.appendChild(ref.$el);
        }
        return ref.init(refState);
    }
    /**
     * Destroys and detaches a specific child component by its reference `id` (as set in `setRef`).
     *
     * @param {string} id Child component reference id
     * @param {boolean} [detach=false] Remove the child component root element from the DOM
     * @returns {Promise}
     */
    async destroyRef(id, detach = false) {
        const { $refsStore } = this;
        const ref = $refsStore.get(id);
        if (!ref) {
            throw new Error(`Child component "${id}" not found.`);
        }
        $refsStore.delete(id);
        delete this.$refs[id];
        if (!detach) {
            return ref.destroy();
        }
        return ref.destroy().then(() => {
            if (ref.$el) {
                ref.$el.remove();
            }
        });
    }
    /**
     * Calls `.destroy()` on every child references and detaches them from the parent component.
     *
     * !> This is an async method returning a promise
     *
     * @returns {Promise}
     */
    async destroyRefs() {
        const { $refsStore } = this;
        if ($refsStore.size === 0) {
            // exit early!
            return Promise.resolve();
        }
        try {
            const result = await Promise.all([...$refsStore.values()].map((ref) => ref.destroy()));
            this.$refs = {};
            $refsStore.clear();
            return result;
        }
        catch (e) {
            this.$warn('An error occurred while destroy the component child components', e);
            return Promise.reject(e);
        }
    }
    /**
     * ```js
     * destroy()
     * ```
     *
     * Detaches DOM events, instance's events and destroys all references as well.
     *
     * > **Lifecycle**
     * >
     * > | stage    | hooks     |
     * > |----------| --------- |
     * > | `destroy` | `beforeDestroy` |
     *
     *
     * !> This is an async method returning a promise
     *
     * @returns {Promise}
     */
    async destroy() {
        await this.beforeDestroy();
        this.removeListeners();
        if (this.$parent) {
            // clear the reference to the parent component
            this.$parent = undefined;
        }
        this.off();
        if (this.$el) {
            this.$el.removeAttribute(Component.UID_DATA_ATTR); // eslint-disable-line no-console
        }
        try {
            await this.destroyRefs();
            this.$active = false;
            return Promise.resolve();
        }
        catch (e) {
            this.$warn('destroy error: ', e);
            return Promise.reject(e);
        }
    }
}
/**
 * ```js
 * Component.UID_DATA_ATTR
 * ```
 * Component root element attribute marker.
 *
 * @static
 * @returns {object}
 */
Component.UID_DATA_ATTR = 'data-cid';
/**
 * Marks yuzu components
 *
 * @static
 */
Component.YUZU_COMPONENT = true;

/**
 * `DetachedComponent` is a special kind of `Component` that does not have a root element (`$el`).
 *
 * @class
 * @see [Component](/packages/yuzu/api/component)
 * @returns {DetachedComponent}
 */
class DetachedComponent extends Component {
    constructor() {
        super(...arguments);
        this.detached = true;
    }
}

/**
 * Element array Iterator.
 *
 * The first argument can be:
 *
 * - a CSS selector. Element will be matched in the context of the passed-in parent's [`$el`](/packages/yuzu/api/component#instance-properties) property.
 * - an array of elements
 * - a function receiving the parent's root element and returning an array of element.
 *
 * The second parameter is an iterator function receiving the current element and its index as arguments.
 *
 * Returns a function that accepts a parent component as first argument and iterates the iterator function over an array of DOM elements.
 *
 * @param {string|Element[]} selector Elements CSS selector, array of elements or a function returning an array of elements.
 * @param {function} fn Iterator function
 * @example
 * import { Children } from 'yuzu';
 *
 * const parent = new ParentComponent().mount('#list');
 * const iterator = (el, i) => new ChildComponent(el, { index: i });
 *
 * const childComponentArray = Children('.items', iterator)(parent);
 *
 * // same as
 * // Children((el) => Array.from(el.querySelectorAll('.items')), iterator)(parent);
 *
 * // same as
 * // const els = Array.from(el.querySelectorAll('.items'))
 * // Children(els , iterator)(parent);
 */
const Children = (selector, fn) => (ctx) => {
    let els;
    if (typeof selector === 'string') {
        els = qsa(selector, ctx.$el);
    }
    else if (typeof selector === 'function') {
        els = selector(ctx.$el);
    }
    else if (Array.isArray(selector)) {
        els = selector;
    }
    else {
        throw new TypeError('Invalid selector argument');
    }
    return els.map(fn);
};

let childRefIdx = 0;
/**
 * `mount` is an helper function to setup trees of components in a functional way.
 *
 * Returns a mount function which in turn accepts a parent component. If present, the returning instance will be attached as reference to the parent component.
 *
 * Child components listed in the `children` parameter will be automatically set as references in the component (uses: [`Component#setRef`](/packages/yuzu/api/component#setref))
 *
 * See the guide on **[functional composition](/packages/yuzu/#functional-composition)** for implementation examples.
 *
 * @param {Component} ComponentConstructor A component constructor (either created by extending `Component` or by [`extend`](/packages/yuzu/api/extend))
 * @param {HTMLElement|string|null} el A mount DOM node (either a CSS selector string or a DOM element). Can be `null` if the component is a [detached component](/packages/yuzu/#detached-components)
 * @param {function[]|function} [children] Child components. Either an array of `mount` functions or a function returning an array of mount functions
 * @param {object} [props] Mount props
 * @param {string} [props.id] Optional component id (used to create a reference onto the parent component)
 * @param {object} [props.state] Component initial state
 * @param {*} [props.*] every other property will be passed as instance option
 * @return {function}
 */
function mount(ComponentConstructor, el, props = {}, children) {
    const _a = props || {}, { state = {}, id, on = {} } = _a, options = __rest(_a, ["state", "id", "on"]);
    const component = new ComponentConstructor(options);
    return function mounter(ctx) {
        if (!component.detached) {
            const root = typeof el === 'string' ? qs(el, ctx && ctx.$el) : el;
            if (isElement(root)) {
                component.mount(root, ctx ? null : state);
            }
        }
        else if (component.detached && !ctx) {
            component.init(state);
        }
        if (ctx) {
            ctx.setRef({
                component,
                id: id || `ref__${++childRefIdx}`,
                on: evaluate(on, ctx),
            }, state);
        }
        const childrenList = Array.isArray(children) ? children : [];
        if (typeof children === 'function') {
            childrenList.push(...children(component));
        }
        for (let i = 0, l = childrenList.length; i < l; i += 1) {
            const child = childrenList[i];
            child(component);
        }
        return component;
    };
}

/**
 * Utility method to create new Component classes in environments that don't support ES6 `class`es.
 *
 * @param {Component} parent Component class to extend
 * @param {object} props New component default properties and methods
 * @example
 * // UMD environment
 * const Text = YZ.extend(YZ.Component, {
 *  created: function () {
 *    this.state = { body:  '' };
 *  },
 *  mounted: function () {
 *    this.$el.innerHTML = this.state.body;
 *  }
 * });
 */
const extend = (parent, props = {}) => {
    const child = props.hasOwnProperty('constructor')
        ? props.constructor
        : function ChildConstructor(options) {
            return parent.call(this, options);
        };
    // https://github.com/mridgway/hoist-non-react-statics/blob/master/index.js#L51
    const keys = Object.getOwnPropertyNames(parent);
    for (let i = 0, l = keys.length; i < l; i += 1) {
        const key = keys[i];
        const descriptor = Object.getOwnPropertyDescriptor(parent, key);
        try {
            // Avoid failures from read-only properties
            Object.defineProperty(child, key, descriptor);
        }
        catch (e) { }
    }
    child.prototype = Object.assign(Object.create(parent.prototype), props);
    child.prototype.constructor = child;
    child.__super__ = parent.prototype;
    return child;
};

/**
 * `devtools` is an helper function that will expose the instance of a Component in a `$yuzu` property attached to its root DOM element.
 *
 * It will also extend every Component instance with some useful methods.
 *
 * !> To improve performance in production, this property and its methods will be available just when `process.env.NODE_ENV !== 'production'`.
 *
 * To initialize the devtools copy the following code into your entry point:
 *
 * ```js
 * import { Component, devtools } from 'yuzu';
 *
 * devtools(Component);
 * ```
 */
let devtools;
if (process.env.NODE_ENV !== 'production') {
    /**
     *
     * @private
     * @param {string} label Logger name
     */
    const createStateLogger = (label) => {
        const $listeners = new Map();
        /**
         * @namespace logger
         * @private
         */
        return {
            label,
            /**
             * Subscribes to a specific event and logs its arguments when emitted. Returns an unsubscribe function.
             *
             * @private
             * @see logger#log
             * @param {Component} instance Target instance
             * @param {string} [event='change:*'] Event to subscribe to
             * @returns {function}
             */
            subscribe(instance, event = 'change:*') {
                const key = event === 'change:*' ? 'change' : event;
                const listener = this.log.bind(this, key);
                if ($listeners.has(event)) {
                    console.warn(`[yuzu-devtools:${this.label}]Already listening for "${event}" on ${instance.$uid}`);
                }
                instance.on(event, listener);
                $listeners.set(event, listener);
                return () => {
                    this.unsubscribe(instance, event);
                };
            },
            /**
             * Unsubscribes the logger from a specific event.
             *
             * @private
             * @param {Component} instance Target instance
             * @param {string} [event='change:*'] Event to subscribe to
             */
            unsubscribe(instance, event = 'change:*') {
                if ($listeners.has(event)) {
                    instance.off(event, $listeners.get(event));
                    $listeners.delete(event);
                }
            },
            /**
             * Unsubscribes the logger from every event.
             *
             * @private
             * @param {Component} instance Target instance
             */
            unsubscribeAll(instance) {
                $listeners.forEach((listener, event) => {
                    instance.off(event, listener);
                });
                $listeners.clear();
            },
            /**
             * Logs formatted data.
             *
             * @private
             * @param {string} msg Log message
             * @param {*} next The current value
             * @param {*} [prev] An optional previous value
             * @param {*[]} args Additional arguments
             */
            log(msg, next, prev, args) {
                if (process.env.NODE_ENV !== 'production') {
                    const head = [
                        `%c${label}: %c${msg}`,
                        'color: gray; font-weight: lighter',
                        'color: green; font-weight: bolder',
                    ];
                    if (args && args.length > 0) {
                        head.push(...args);
                    }
                    console.groupCollapsed(...head);
                    if (prev !== undefined) {
                        console.log('%cprev state', 'color: gray; font-weight: bolder', prev);
                        console.log('%cnext state', 'color: green; font-weight: bolder', next);
                    }
                    else {
                        console.log('%cinitial state', 'color: gray; font-weight: bolder', next);
                    }
                    console.groupEnd();
                }
            },
        };
    };
    /* eslint-disable no-param-reassign */
    devtools = (ComponentClass) => {
        /**
         * Methods and properties added by devtools
         *
         * @name Component
         */
        const proto = ComponentClass.prototype;
        const { mount, init, destroy } = proto;
        // function refTree(root: Component, tree: IObject) {
        //   if (root.$refsStore.size > 0) {
        //     root.$refsStore.forEach((ref, name) => {
        //       tree[name] = Object.create(null);
        //       refTree(ref, tree[name]);
        //     });
        //   }
        //   Object.defineProperty(tree, '$self', {
        //     enumerable: false,
        //     value: Object.assign(Object.create(null), {
        //       $raw: root,
        //       state: Object.assign(Object.create(null), root.state),
        //     }),
        //   });
        //   return tree;
        // }
        proto.mount = function mountDev(...args) {
            mount.call(this, ...args);
            /**
             * A reference to a Component instance attached to its root element.
             *
             * @name $el.$yuzu
             * @type Component
             * @memberof Component
             */
            Object.defineProperty(this.$el, '$yuzu', {
                value: this,
                configurable: true,
            });
            // Object.defineProperty(this, '$$getTree', {
            //   enumerable: false,
            //   writable: false,
            //   value() {
            //     const tree = Object.create(null);
            //     return refTree(this, tree);
            //   },
            // });
            return this;
        };
        proto.destroy = function destroyDev() {
            const $el = this.$el;
            if ($el && $el.$yuzu) {
                delete $el.$yuzu;
            }
            delete this.$$logStart;
            delete this.$$logEnd;
            return destroy.call(this);
        };
        proto.init = function initDev(state) {
            Object.defineProperties(this, {
                /**
                 * ```js
                 * $$logStart([label], [event])
                 * ```
                 *
                 * Will initialize an event logger with a custom label. By default will automatically log any change to the instance `state` property.
                 *
                 * ?> Note that event loggers will be not available until the instance is initialized
                 *
                 * @memberof Component
                 * @param {string} [label] Log label. If not defined will fallback to: A `debugLabel` option, a static `displayName` property defined on the constructor, the component name
                 * @param {string|boolean} [listen="change:*"] Event to listen for or `false`.
                 */
                $$logStart: {
                    enumerable: false,
                    configurable: true,
                    value(label = null, listen = 'change:*') {
                        const name = label ||
                            this.options.debugLabel ||
                            `${this.constructor.displayName ||
                                this.constructor.name ||
                                'Component'}#${this.$uid}`;
                        if (!this.$$logger) {
                            this.$$logger = createStateLogger(name);
                        }
                        if (listen) {
                            this.$$logger.subscribe(this, listen);
                        }
                    },
                },
                /**
                 * ```js
                 * $$logEnd([event])
                 * ```
                 *
                 * Will stop an event logger either for a specific event or for every event.
                 *
                 * @memberof Component
                 * @param {string} [event] Event to unsubscribe for. If not provided the logger will be completely stopped and removed
                 */
                $$logEnd: {
                    enumerable: false,
                    configurable: true,
                    value(event) {
                        if (this.$$logger) {
                            if (event) {
                                this.$$logger.unsubscribe(this, event);
                                return;
                            }
                            this.$$logger.unsubscribeAll(this);
                            this.$$logger = undefined;
                        }
                    },
                },
            });
            return init.call(this, state);
        };
        // @ts-ignore: Devtools Hooks
        if (window.__YUZU_DEVTOOLS_GLOBAL_HOOK__) {
            // @ts-ignore
            window.__YUZU_DEVTOOLS_GLOBAL_HOOK__.init(Component);
        }
    };
    /* eslint-enable no-param-reassign */
}
else {
    devtools = () => undefined;
}

export { Children, Component, DetachedComponent, devtools, extend, mount };
//# sourceMappingURL=index.next.js.map
